import logging

import click
import numpy as np
from tifffile import TiffFile

from dbimage.io import write_array_to_fpath


def arraydata_from_tif_fpath(tif_fpath):

    with TiffFile(tif_fpath) as tif:
        rawdata = tif.asarray()

    arraydata = np.transpose(rawdata, (1, 2, 0))

    return arraydata


@click.command()
@click.argument('input_tif_fpath')
@click.argument('output_dbim_fpath')
def convert_tif_image(input_tif_fpath, output_dbim_fpath):

    logging.basicConfig(level=logging.DEBUG)

    arraydata = arraydata_from_tif_fpath(input_tif_fpath)
    logging.info(f"Loaded TIFF with shape {arraydata.shape}, dtype {arraydata.dtype}")

    write_array_to_fpath(output_dbim_fpath, arraydata)



import logging

import click

from dbimage.io import read_array_info_from_fpath, read_dbim_from_fpath


@click.command()
@click.argument('dbim_fpath')
def dbim_info(dbim_fpath):

    logging.basicConfig(level=logging.DEBUG)

    array_info = read_array_info_from_fpath(dbim_fpath)
    print(array_info)
    # print(f"Read array with shape {sa.shape}, dtype {sa.dtype}")



import struct

import numpy as np

from dbimage import (
    HDR_FORMAT,
    bytes_to_array_info,
    compressed_bytes_to_shaped_array,
    arraydata_to_compressed_bytes,
    header_bytes_from_array
)


def read_array_info_from_fpath(fpath):

    with open(fpath, "rb") as fh:
        hdr_bytes = fh.read(struct.calcsize(HDR_FORMAT))

    array_info = bytes_to_array_info(hdr_bytes)

    return array_info


def read_dbim_from_fpath(fpath):

    with open(fpath, 'rb') as fh:
        hdr_bytes = fh.read(struct.calcsize(HDR_FORMAT))
        cbytes = fh.read()

    array_info = bytes_to_array_info(hdr_bytes)

    shaped_array = compressed_bytes_to_shaped_array(cbytes, array_info)

    return shaped_array


def write_array_to_fpath(fpath, array):

    cbytes = arraydata_to_compressed_bytes(array)
    hdr_bytes = header_bytes_from_array(array, cbytes)

    with open(fpath, 'wb') as fh:
        fh.write(hdr_bytes)
        fh.write(cbytes)



from setuptools import setup

setup(name='dbimage',
      version='0.1.0',
      description='dtoolbioimage image format',
      packages=['dbimage'],
      url='https://github.com/JIC-Image-Analysis/dbimage',
      author='Matthew Hartley',
      author_email='Matthew.Hartley@jic.ac.uk',
      license='MIT',
      install_requires=[
	      "click",
        "numpy",
        "blosc",
        "tifffile"
      ],
      entry_points='''
        [console_scripts]
        tiff2dbim=dbimage.convert:convert_tif_image
        dbim_info=dbimage.info:dbim_info
      ''',
      zip_safe=False)



import sys
import struct
import logging

import blosc
import numpy as np



DTYPE_CODE_LOOKUP = {
    np.dtype('uint8'): 0
}

DTYPE_REVERSE_LOOKUP = {
    0: np.dtype('uint8')
}

HDR_FORMAT = "iQQ" + "LLL" + "LLL"


def array_info_from_array(array):

    strides = array.__array_interface__['strides']
    if strides is None:
        strides = 0, 0, 0

    array_info = dict(
        dtype=array.dtype,
        size=array.size,
        strides=strides,
        dim=array.shape
    )

    return array_info


def array_info_to_hdr_values(array_info):

    dtype = array_info['dtype']
    try:
        dtype_code = DTYPE_CODE_LOOKUP[dtype]
    except KeyError:
        print(f"Can't handle type {dtype}")
        sys.exit(2)

    hdr_szinfo = (dtype_code, array_info['size'], array_info['packed_size'])
    hdr_values = hdr_szinfo + array_info['dim'] + array_info['strides']

    return hdr_values


def header_bytes_from_array(array, cbytes):

    array_info = array_info_from_array(array)
    array_info['packed_size'] = len(cbytes)
    logging.debug(f"Packing with array_info={array_info}")

    hdr_values = array_info_to_hdr_values(array_info)
    logging.debug(f"Packing header values {hdr_values}")
    hdr_packed = struct.pack(HDR_FORMAT, *hdr_values)

    return hdr_packed


def bytes_to_array_info(hdr_bytes):
    hdr_values = struct.unpack(HDR_FORMAT, hdr_bytes)
    logging.debug(f"Read header values: {hdr_values}")

    dtype_code, size, packed_size, *rest = hdr_values
    rdim, cdim, zdim, *strides = rest


    array_info = dict(
        dtype=DTYPE_REVERSE_LOOKUP[dtype_code],
        size=size,
        packed_size=packed_size,
        dim=(rdim, cdim, zdim),
        strides=tuple(strides)
    )

    return array_info


def arraydata_to_compressed_bytes(arraydata):

    cbytes = blosc.compress_ptr(
        arraydata.__array_interface__['data'][0],
        arraydata.size,
        arraydata.dtype.itemsize,
        cname='zlib',
        shuffle=blosc.SHUFFLE
    )

    return cbytes


def compressed_bytes_to_arraydata(cbytes, size, dtype):

    arraydata = np.empty(size, dtype=dtype)
    blosc.decompress_ptr(cbytes, arraydata.__array_interface__['data'][0])

    return arraydata


def compressed_bytes_to_shaped_array(cbytes, array_info):

    size = array_info['size']
    dtype = array_info['dtype']
    arraydata = compressed_bytes_to_arraydata(cbytes, size, dtype)

    strides = array_info['strides']
    dim = array_info['dim']
    if strides == (0, 0, 0):
        shaped_array = np.reshape(arraydata, dim)
    else:
        shaped_array = np.lib.stride_tricks.as_strided(arraydata, dim, strides, writeable=False)

    return shaped_array


